function ownKeys(object, enumerableOnly) { var keys = Object.keys(object); if (Object.getOwnPropertySymbols) { var symbols = Object.getOwnPropertySymbols(object); if (enumerableOnly) { symbols = symbols.filter(function (sym) { return Object.getOwnPropertyDescriptor(object, sym).enumerable; }); } keys.push.apply(keys, symbols); } return keys; }

function _objectSpread(target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i] != null ? arguments[i] : {}; if (i % 2) { ownKeys(Object(source), true).forEach(function (key) { _defineProperty(target, key, source[key]); }); } else if (Object.getOwnPropertyDescriptors) { Object.defineProperties(target, Object.getOwnPropertyDescriptors(source)); } else { ownKeys(Object(source)).forEach(function (key) { Object.defineProperty(target, key, Object.getOwnPropertyDescriptor(source, key)); }); } } return target; }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

import React, { useEffect, useState, useRef } from 'react';
import { isPlainObject, isFunction } from 'lodash';

var InitialProxyActions = () => proxyActions({});

function proxyActions(actions) {
  return new Proxy(actions, {
    get(trapTarget, key, receiver) {
      if (!(key in receiver)) Reflect.set(trapTarget, key, () => {}, receiver); // TODO: 抛出提示

      return Reflect.get(trapTarget, key, receiver);
    }

  });
}

function reduceMutation(curMutation, getMutate) {
  return Object.keys(curMutation).reduce((nextMutation, type) => Object.assign({}, nextMutation, {
    [type]: getMutate(type)
  }), {});
}

export default function createContext(state, reducer) {
  var InitialState = isPlainObject(state) ? _objectSpread({}, state) : {};
  var namespaces = Object.keys(InitialState);
  var mutate = getMutateByNamespaces(namespaces);
  var configReducer = isFunction(reducer) ? reducer : () => InitialState;
  var InitialMutation = namespaces.reduce((mutation, type) => Object.assign(mutation, {
    [type]: () => InitialState
  }), {});
  var InitialContextData = {
    context: InitialState,
    mutation: InitialMutation,
    useActions: InitialProxyActions
  };
  var AppContext = /*#__PURE__*/React.createContext(InitialContextData);

  function AppProvider(props) {
    var {
      config = {},
      children
    } = props;
    var actionsRef = useRef(null); // 让useActions不可变

    var stateRef = useRef(InitialState); // 给actions的闭包用

    var [isInit, setIsInit] = useState(false);
    var [contextData, setContextData] = useState(InitialContextData);
    useEffect(() => {
      if (isInit) return;
      var context = configReducer(config)(stateRef.current); // TODO: catch 异常并给出来自包的提示

      stateRef.current = context;
      setContextData(_ref => {
        var {
          mutation: curMutation,
          useActions: curUseActions
        } = _ref;
        var mutation = reduceMutation(curMutation, type => (payload, isSetState) => isFunction(payload) && setContextData(curContextData => {
          var nextContext = mutate({
            type,
            payload
          }, stateRef.current, {
            isSetState
          });
          stateRef.current = nextContext;
          return _objectSpread(_objectSpread({}, curContextData), {}, {
            context: nextContext
          });
        }));

        var useActions = getActions => {
          if (actionsRef.current) return actionsRef.current;
          return actionsRef.current = isFunction(getActions) ? proxyActions(getActions(mutation, stateRef)) : curUseActions;
        };

        return {
          context,
          mutation,
          useActions
        };
      });
      setIsInit(true);
    }, [isInit, config, stateRef, actionsRef]);
    return /*#__PURE__*/React.createElement(AppContext.Provider, {
      value: contextData
    }, children);
  }

  return {
    AppProvider,
    AppConsumer: AppContext.Consumer
  };
}

function getMutateByNamespaces(namespaces) {
  return function mutate(action, state, _ref2) {
    var {
      isSetState
    } = _ref2;
    var {
      type,
      payload
    } = action;
    var namespace = namespaces.find(name => name === type);
    if (!namespace) return state;
    return Object.assign({}, state, {
      [type]: isSetState ? setState(payload, state[type]) : mergeState(payload, state[type])
    });
  };
}

function setState(payload, state) {
  return Object.assign({}, state, payload(state));
}

function mergeState(payload, state) {
  var nextState = payload(state);
  return Object.assign({}, state, Object.keys(nextState).reduce((newState, key) => {
    var subState = state[key];

    if (isPlainObject(subState)) {
      Object.assign(newState, {
        [key]: Object.assign({}, subState, nextState[key])
      });
    } else {
      Object.assign(newState, {
        [key]: nextState[key]
      });
    }

    return newState;
  }, {}));
}